from collections import defaultdict
from datetime import datetime

from marshmallow import ValidationError


class Validateable(object):
    """Mixin for objects supporting validation through marshmallow schema."""

    def __init__(self):
        self._errors = None

    @property
    def errors(self):
        """Errors dictionary from last validation or None"""
        if not self._errors:
            return None

        return dict(self._errors)

    @property
    def is_valid(self):
        """Calls :meth:`validate` and returns validation result."""

        self.validate()
        return not self._errors

    @property
    def _schema():
        """
        marshmallow Schema used to validate self.

        Note:
            It is allowed to return None in which case validation will report
            schema error.
        """
        raise NotImplementedError(
            "class {} is missing implementation of self._schema!".
            format(self.__class__.__name__)
        )

    def validate(self):
        """
        Validates instance using schema.

        Returns:
            dict: Errors dictionary.
        """
        self._errors = defaultdict(list)
        schema = self._schema
        if schema:
            try:
                json_str, dump_errors = schema.dumps(self)
                # Ignoring dump_errors because they don't contain all
                # validations - this is marshmallows' by-design feature.
                # Instead, we'll try to load it now and use that as errors dict
                data, self._errors = schema.loads(json_str)

            except ValidationError as e:
                for k, v in e.message.items():
                    self._errors[k] = v

            except Exception as e:
                self._errors[self.__class__.__name__].append(str(e))
        else:
            self._errors['{} schema'.format(self.__class__.__name__)
                        ].append('Invalid message_type!')

        return self._errors


class Representable(object):
    """
    Mixin provides generic __repr__ implementation for value objects (for
    example: models).

    - prints only public attributes
    - date values are printed as ISO8601
    """

    _ONLY_REPR_ATTRIBUTES = None
    _SKIP_REPR_ATTRIBUTES = None

    def _repr_it(self, name, value):
        return "{}={}".format(
            name, repr(getattr(value, "isoformat", lambda: value)())
        )

    @property
    def _attrs_to_repr(self):
        return sorted([
            attr for attr in (
                set(self._ONLY_REPR_ATTRIBUTES or [])
                or (
                    set(vars(self).keys()) -
                    set(self._SKIP_REPR_ATTRIBUTES or [])
                )
            )
            if not attr.startswith('_')
        ])

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__, ", ".join([
                self._repr_it(attr, getattr(self, attr))
                for attr in self._attrs_to_repr
            ])
        )



import six


def is_blank(obj):
    """
    True if line is empty string, None, string that contains only spaces and
    space like characters, or line is iterable that contains only these kinds
    of strings/objects
    """
    if not obj:
        return True

    if isinstance(obj, six.string_types):
        return not obj.strip()

    retv = False
    try:
        retv = all(is_blank(x) for x in obj)
    except TypeError:
        # ... not iterable
        retv = False

    return retv



from seveno_pyutil import dict_utilities


class describe_inverted(object):
    def it_inverts_dict_keys_and_values(self):
        d = {'a': 1, 'b': 2}

        assert dict_utilities.inverted(d) == {1: 'a', 2: 'b'}



import errno
import os
import shutil

from .string_utilities import is_blank


def file_checksum(file_path, hash_callable):
    """Given path of the file and hash function, calculates file digest"""
    if os.path.isfile(file_path) and callable(hash_callable):
        hash_obj = hash_callable()
        with open(file_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_obj.update(chunk)
        return hash_obj.hexdigest()

    return None


def silent_create_dirs(dir_path):
    """Tries to create directory and silently skips if it exists."""
    try:
        os.makedirs(os.path.abspath(dir_path))
    except OSError as exception:
        # We don't care if dir already exists
        if (
            exception.errno != errno.EEXIST
            or (
                exception.errno == errno.EEXIST and not os.path.isdir(
                    os.path.abspath(dir_path)
                )
            )
        ):
            raise


def abspath_if_relative(relative_path, relative_to):
    """Creates absolute path from relative, but places it under other path.

    Example:
        >>> abspath_if_relative('foo/bar/baz', relative_to='/tmp')
        '/tmp/foo/bar/baz'
    """
    retv = relative_path

    if not is_blank(relative_path):
        if not is_blank(relative_to):
            if not os.path.isabs(relative_path):
                retv = os.path.abspath(
                    os.path.join(relative_to, relative_path)
                )
        else:
            retv = os.path.abspath(relative_path)

    return retv


def move_and_create_dest(src_path, dst_dir):
    """
    Moves ``src_path`` to ``dst_dir`` directory.

    Expects ``dst_dir`` to be directory and if it doesn't exits, tries to create
    it.
    """
    silent_create_dirs(dst_dir)
    shutil.move(src_path, dst_dir)
    return os.path.join(dst_dir, os.path.basename(src_path))


def silent_remove(file_or_dir_path):
    """
    Deletes file or directory (even if not empty). Doesn't rise if it doesn't
    exist.
    """

    if is_blank(file_or_dir_path):
        return None

    try:
        if (os.path.isdir(file_or_dir_path)):
            shutil.rmtree(file_or_dir_path)
        else:
            os.remove(file_or_dir_path)

    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def switch_extension(file_path, new_extension):
    if is_blank(new_extension):
        return file_path

    if not new_extension.startswith('.'):
        new_extension = '.' + new_extension

    return os.path.join(
        os.path.dirname(file_path),
        os.path.splitext(
            os.path.basename(file_path)
        )[0] + new_extension
    )



import itertools


def in_batches(iterable, of_size=1):
    """
    Generator that yields generator slices of iterable.

    Since it is elegant and working flawlessly, it is shameles C/P from
    https://stackoverflow.com/a/8998040/2219433

    Warning:
        Each returned batch should be completely consumed before next batch
        is yielded. See example below to better understand what that means.

    Example:

        >>> from seveno_pyutil import in_batches
        >>> g = (o for o in range(10))
        >>> for batch in in_batches(g, of_size=3):
        ...     print(list(batch))
        ...
        [0, 1, 2]
        [3, 4, 5]
        [6, 7, 8]
        [9]
        >>> g = list(range(10))
        >>> for batch in in_batches(g, of_size=3):
        ...     print(list(batch))
        ...
        [0, 1, 2]
        [3, 4, 5]
        [6, 7, 8]
        [9]
        >>> # And now a demo what happens if don't consume whole batch before
        >>> # yielding another one
        >>> g = list(range(10))
        >>> g
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        >>> for batch in in_batches(g, of_size=3):
        ...     print( [next(batch), next(batch)] )
        ...
        [0, 1]
        [2, 3]
        [4, 5]
        [6, 7]
        [8, 9]
    """
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, of_size)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)
